/*************************************************************************/
/*									 */
/*	Evaluation of a test on a discrete valued attribute		 */
/*      ---------------------------------------------------		 */
/*									 */
/*************************************************************************/


#include "buildex.i"


/*************************************************************************/
/*									 */
/*  Set Info[] and Gain[] for discrete partition of items Fp to Lp	 */
/*									 */
/*************************************************************************/


void EvalDiscreteAtt(Att, Fp, Lp, Items)
/*  ---------------  */ 
    Attribute Att;
    ItemNo Fp, Lp; 
    ItemCount Items;
{ 
    ItemCount KnownItems;
    float DiscrKnownBaseInfo(), ComputeGain(), MyMaskedInfo(), 
      MySplitInfo();
    void ComputeFrequencies(),PrintDistribution();

    ItemNo i; short n;

    int length = MaxClass+1;
    ClassList tempclassmask,classmask,existingclasses,classes;
    ClassList ClassesOf();
    struct Classtree **queue, **children, *class;
    
    float tempgain=-Epsilon, finalgain = -Epsilon;

    queue           = NewQueueOfClassTreeElems(length);
    children        = NewQueueOfClassTreeElems(length);
    classmask       = NewClassList(length);
    tempclassmask   = NewClassList(length);
    existingclasses = NewClassList(length);

    ForEach(i,Fp,Lp) {
      classes = ClassesOf(Item[i]);
      ForEach(n,0,MaxClass) {
	if(classes[n]) {existingclasses[n] = true;}
      }
    }

    if(!GetListOfChildrenIfExist(classTree,queue,existingclasses)) 
      { printf("FormTree: Nothing in classtree!\n"); exit(1); }
    SetBits(classmask,queue,length);

    /* Slightly different for first time through, since classmask bits 
       are already set and we only need to calculate KnownItems once 
    */
    ComputeFrequencies(Att, Fp, Lp, classmask);
    KnownItems = Items - ValFreq[0];
    /*  Special case when no known values of the attribute  */
    if ( Items <= ValFreq[0] )
      {
	Verbosity(2) printf("\tAtt %s: no known values\n", AttName[Att]);
	
	Gain[Att] = -Epsilon;
	Info[Att] = 0.0;
	return;
      }
    finalgain =  ComputeGain( MyMaskedInfo(ClassFreq,0,MaxClass,
			    classmask), UnknownRate[Att], 
                            MaxAttVal[Att], KnownItems, classmask);


    while(!EmptyQueue(queue))
      {
	class = PopFromQueue( queue );
	if(GetListOfChildrenIfExist( class, children, existingclasses ))
	  {
	    CopyClassList( tempclassmask, classmask, length );
	    SetBitsOffAndOn( tempclassmask, class, children, length);

	    ComputeFrequencies(Att, Fp, Lp, tempclassmask);
	    tempgain = ComputeGain( MyMaskedInfo(ClassFreq,0,MaxClass,
			    tempclassmask), UnknownRate[Att], 
                            MaxAttVal[Att], KnownItems, tempclassmask);
	    if(tempgain > finalgain)
	      {
		CopyClassList(classmask,tempclassmask,length);
		finalgain = tempgain;
		AddToQueue(queue,children);
	      }
	    Verbosity(2)
	      {
		printf("\tAtt %s", AttName[Att]);
		Verbosity(3) PrintDistribution(Att, MaxAttVal[Att], true);
		printf("\tinf %.3f, gain %.3f\n", Info[Att], Gain[Att]);
	      }
	  }
      }
    
    Gain[Att] = finalgain;
    /* split info */
    Info[Att] = MySplitInfo(ValFreq, 0, MaxAttVal[Att]);
 

    FreeQueueOfClassTreeElems(queue);
    FreeQueueOfClassTreeElems(children);
    FreeClassList(classmask);
    FreeClassList(tempclassmask);
    FreeClassList(existingclasses);
} 



/*************************************************************************/
/*									 */
/*  Compute frequency tables Freq[][] and ValFreq[] for attribute	 */
/*  Att from items Fp to Lp, and set the UnknownRate for Att		 */
/*									 */
/*************************************************************************/


void ComputeFrequencies(Att, Fp, Lp, Classmask)
/*  ------------------  */
    Attribute Att;
    ItemNo Fp, Lp;
    ClassList Classmask;
{
    Description Case; 
    ItemCount CountItems();
    ItemNo p;
    int n, NoClasses;
    float weight;
    DiscrValue attvalue;

    ClassList classes;

    ClassList ClassesOf();
    void ResetFreq();
    int NumberOfMaskedClasses();

    ResetFreq(MaxAttVal[Att]);


    /*  Determine the frequency of each class amongst cases
	with each possible value for the given attribute  */

    ForEach(p, Fp, Lp)
    { 
	Case = Item[p];
	NoClasses = NumberOfMaskedClasses( Case, Classmask, MaxClass+1 );
	/*	normalisedWeight = Weight[p]/NoClasses; */
	classes = ClassesOf(Case);
	attvalue = DVal(Case,Att);
	weight = Weight[p];
	ForEach(n, 0, MaxClass)
	  {
	    if(classes[n] && Classmask[n])
	      {
		Freq[ attvalue ][ n ] += weight;
		ClassFreq[ n ] += weight;
	      }
	  }
	/*  Determine the frequency of each possible value for the
	    given attribute  */
	ValFreq[ attvalue ] += weight;

    } 
    
    /*  Set the rate of unknown values of the attribute  */
 
    UnknownRate[Att] = ValFreq[0] / CountItems(Fp, Lp);
}



/*************************************************************************/
/*									 */
/*  Return the base info for items with known values of a discrete	 */
/*  attribute, using the frequency table Freq[][]			 */
/*	 								 */
/*************************************************************************/


float DiscrKnownBaseInfo(KnownItems, MaxVal)
/*    ------------------  */
    DiscrValue MaxVal;
    ItemCount KnownItems;
{
    ClassNo c;
    ItemCount ClassCount;
    double Sum=0;
    DiscrValue v;

    /* this type conversion is needed for egcs */
    int I_MaxVal = MaxVal;

    ForEach(c, 0, MaxClass)
    {
	ClassCount = 0;
	ForEach(v, 1, I_MaxVal)
	{
	    ClassCount += Freq[v][c];
	}
	Sum += ClassCount * Log(ClassCount);
    }

    return (KnownItems * Log(KnownItems) - Sum) / KnownItems;
}



/*************************************************************************/
/*									 */
/*  Construct and return a node for a test on a discrete attribute	 */
/*									 */
/*************************************************************************/


void DiscreteTest(Node, Att)
/*  ----------  */
    Tree Node;
    Attribute Att;
{
    ItemCount CountItems();

    Sprout(Node, MaxAttVal[Att]);

    Node->NodeType	= BrDiscr;
    Node->Tested	= Att;
    Node->Errors	= 0;
}

